# -*- coding: utf-8 -*-
from stdlib_list import stdlib_list
from z3c.dependencychecker.dotted_name import DottedName
import logging
import sys


logger = logging.getLogger(__name__)


class ImportsDatabase(object):
    """Store all imports and requirements of a package

    Use that information to extract useful data out of it,
    like requirements unused, etc.
    """

    def __init__(self):
        self._requirements = set()
        self._extras_requirements = {}
        self.imports_used = []
        self.user_mappings = {}
        self.reverse_user_mappings = {}
        self.ignored_packages = set()
        self.own_dotted_name = None

    def add_requirements(self, requirements):
        self._requirements = set([
            requirement for requirement in requirements
        ])

    def add_extra_requirements(self, extra_name, dotted_names):
        # A bit of extra work needs to be done as pkg_resources API returns
        # all common requirements as well as the extras when asking for an
        # extra requirements.
        only_extra_dotted_names = self._filter_duplicates(dotted_names)
        if extra_name in self._extras_requirements.keys():
            self._extras_requirements[extra_name].update(
                only_extra_dotted_names,
            )

            logger.warn(
                'extra requirement "%s" is declared twice in setup.py',
                extra_name,
            )
        else:
            self._extras_requirements[extra_name] = only_extra_dotted_names

    def _filter_duplicates(self, imports):
        """Return all items in imports that are not a requirement already"""
        all_imports = set([dotted_name for dotted_name in imports])
        filtered = all_imports - self._requirements
        return filtered

    def add_imports(self, imports):
        filters = (
            self._filter_out_known_packages,
            self._filter_out_own_package,
            self._filter_out_python_standard_library,
        )
        for single_import in imports:
            logger.debug('    Import found: %s', single_import.name)

            unknown_import = self._apply_filters([single_import, ], filters)
            if unknown_import:
                self.imports_used.append(single_import)
            else:
                logger.debug('    Import ignored: %s', single_import.name)

    def add_user_mapping(self, package_name, provided_names):
        package = DottedName(package_name)
        packages_provided = [DottedName(name) for name in provided_names]

        if package not in self._all_requirements():
            logger.info(
                'Ignoring package %s as is not a dependency of the '
                'package being analyzed',
                package,
            )
            return

        self.user_mappings[package] = set(packages_provided)

        for single_package in packages_provided:
            self.reverse_user_mappings[single_package] = package

    def add_ignored_packages(self, packages):
        self.ignored_packages = set([
            DottedName(package)
            for package in packages
        ])

    def _all_requirements(self):
        all_requirements = self._requirements.copy()
        for extra in self._extras_requirements:
            all_requirements.update(self._extras_requirements[extra])

        return all_requirements

    def get_missing_imports(self):
        filters = (
            self._filter_out_testing_imports,
            self._filter_out_requirements,
            self._filter_out_ignored_imports,
        )
        missing = self._apply_filters(self.imports_used, filters)
        unique_imports = self._get_unique_imports(imports_list=missing)
        result = self._filter_user_mappings(unique_imports)
        return result

    def get_missing_test_imports(self):
        filters = (
            self._filter_out_only_testing_imports,
            self._filter_out_requirements,
            self._filter_out_test_requirements,
            self._filter_out_ignored_imports,
        )
        missing = self._apply_filters(self.imports_used, filters)
        unique_imports = self._get_unique_imports(imports_list=missing)
        result = self._filter_user_mappings(unique_imports)
        return result

    def get_unneeded_requirements(self):
        all_but_test_requirements = self._requirements.copy()
        for extra in self._extras_requirements:
            all_but_test_requirements.update(self._extras_requirements[extra])
        test_requirements = self._get_test_extra()
        for dotted_name in test_requirements:
            all_but_test_requirements.remove(dotted_name)
        filters = (
            self._filter_out_known_packages,
            self._filter_out_python_standard_library,
            self._filter_out_used_imports,
            self._filter_out_ignored_imports,
        )
        unneeded = self._apply_filters(all_but_test_requirements, filters)
        unique_imports = self._get_unique_imports(imports_list=unneeded)
        return unique_imports

    def requirements_that_should_be_test_requirements(self):
        non_testing_filters = (
            self._filter_out_testing_imports,
        )
        non_testing_imports = self._apply_filters(
            self.imports_used,
            non_testing_filters,
        )
        requirements_not_used = [
            requirement
            for requirement in self._requirements
            if self._discard_if_found_obj_in_list(
                requirement,
                non_testing_imports,
            )
        ]
        testing_filters = (
            self._filter_out_only_testing_imports,
            self._filter_out_own_package,
        )
        testing_imports = self._apply_filters(
            self.imports_used,
            testing_filters,
        )
        should_be_test_requirements = [
            requirement
            for requirement in requirements_not_used
            if not self._discard_if_found_obj_in_list(
                requirement,
                testing_imports,
            )
        ]
        filters = (
            self._filter_out_ignored_imports,
        )
        skip_ignored = self._apply_filters(
            should_be_test_requirements,
            filters,
        )
        unique = self._get_unique_imports(imports_list=skip_ignored)
        return unique

    def get_unneeded_test_requirements(self):
        test_requirements = self._get_test_extra()
        if not test_requirements:
            return []

        filters = (
            self._filter_out_known_packages,
            self._filter_out_python_standard_library,
            self._filter_out_used_imports,
            self._filter_out_ignored_imports,
        )
        unneeded = self._apply_filters(test_requirements, filters)
        unique_imports = self._get_unique_imports(imports_list=unneeded)
        return unique_imports

    @staticmethod
    def _apply_filters(objects, filters):
        """Filter a list of given objects through a list of given filters

        This helps encapsulating the filtering logic here while keeping the
        methods free from this clutter.

        Callees of this method need to provide a list of objects
        (usually a list of DottedNames) that will be processed by the provided
        list of filters.

        A filter, in this context, is a function that takes a DottedName as a
        single argument and returns a boolean value.

        If a given DottedName returns True in all filters,
        it will make it to the final result an be returned back.
        """
        result = []
        for single_object in objects:
            for filter_func in filters:
                if not filter_func(single_object):
                    break
            else:
                result.append(single_object)

        return result

    def _get_unique_imports(self, imports_list=None):
        if imports_list is None:
            imports_list = self.imports_used
        unique_dotted_names = list(set(imports_list))
        sorted_unique_dotted_names = sorted(unique_dotted_names)
        return sorted_unique_dotted_names

    def _filter_out_used_imports(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self.imports_used,
        )

    def _filter_out_ignored_imports(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self.ignored_packages,
        )

    def _filter_out_known_packages(self, dotted_name):
        to_ignore = (
            DottedName('setuptools'),
            DottedName('pkg_resources'),
            DottedName('distribute'),
        )
        return self._discard_if_found_obj_in_list(dotted_name, to_ignore)

    @staticmethod
    def _filter_out_testing_imports(dotted_name):
        return not dotted_name.is_test

    @staticmethod
    def _filter_out_only_testing_imports(dotted_name):
        return dotted_name.is_test

    def _filter_out_own_package(self, dotted_name):
        return dotted_name not in self.own_dotted_name

    def _filter_out_requirements(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self._requirements,
        )

    def _filter_out_test_requirements(self, dotted_name):
        test_requirements = self._get_test_extra()
        if not test_requirements:
            return True

        return self._discard_if_found_obj_in_list(
            dotted_name,
            test_requirements,
        )

    def _filter_out_python_standard_library(self, dotted_name):
        std_library = self._build_std_library()
        return self._discard_if_found_obj_in_list(dotted_name, std_library)

    @staticmethod
    def _discard_if_found_obj_in_list(obj, obj_list):
        """Check if obj is on obj_list and return the opposite

        We are interested in filtering out requirements,
        so if a given object is on a given list,
        that object can be discarded as it's already in the list.
        """
        # this is a generator expression, nothing is computed until it is being
        # called some lines below
        iterable = (
            item
            for item in obj_list
            if obj in item
        )

        # any builtin stops consuming the iterator as soon as an element of the
        # iterable is True
        if any(iterable):
            # if 'obj' is found, the pipeline needs to stop and discard the
            # requirement
            return False

        # 'obj' was not found on the list, means that is fulfilling the
        # pipeline filters so far
        return True

    @staticmethod
    def _build_std_library():
        python_version = sys.version_info
        libraries = stdlib_list(
            '{0}.{1}'.format(
                python_version[0],
                python_version[1],
            )
        )

        fake_std_libraries = [DottedName(x) for x in libraries]
        return fake_std_libraries

    def _get_test_extra(self):
        candidates = ('test', 'tests', )
        for candidate in candidates:
            if candidate in self._extras_requirements:
                return self._extras_requirements[candidate]

        return []

    def _filter_user_mappings(self, dotted_names):
        """Remove dotted names that are in user mappings"""
        result = []
        for single_import in dotted_names:
            for provided_package in self.reverse_user_mappings:
                if single_import in provided_package:
                    logger.debug(
                        'Skip %s as is part of user mapping %s',
                        single_import,
                        self.reverse_user_mappings[provided_package],
                    )
                    break
            else:
                result.append(single_import)

        return result



# -*- coding: utf-8 -*-
from z3c.dependencychecker.dotted_name import DottedName
import mock
import pytest
import sys


def test_minimal():
    obj = DottedName('dotted.name')
    assert obj.name == 'dotted.name'


def test_no_path():
    obj = DottedName('dotted.name')
    assert obj.file_path is None


def test_file_path_given():
    obj = DottedName('dotted.name', file_path='/one/two')
    assert obj.file_path == '/one/two'


def test_requirement_constructor():
    mock_object = mock.MagicMock()
    mock_property = mock.PropertyMock(return_value='my.dotted.name')
    type(mock_object).project_name = mock_property

    obj = DottedName.from_requirement(mock_object)

    assert obj.file_path is None
    assert obj.name == 'my.dotted.name'


def test_requirement_constructor_with_path():
    mock_object = mock.MagicMock()
    mock_property = mock.PropertyMock(return_value='my.dotted.name')
    type(mock_object).project_name = mock_property

    obj = DottedName.from_requirement(mock_object, file_path='/one/two')

    assert obj.file_path == '/one/two'


def test_is_test_import():
    obj = DottedName('plone.app.dexterity', is_test=True)
    assert obj.is_test


def test_not_a_test_import():
    obj = DottedName('plone.app.dexterity')
    assert not obj.is_test


def test_safe_name_no_changes():
    obj = DottedName('hi.there')
    assert obj.name == obj.safe_name


def test_safe_name_lowercase():
    original_name = 'Hi.There'
    obj = DottedName(original_name)
    assert original_name.lower() == obj.safe_name


def test_safe_name_dash_to_underscore():
    original_name = 'hi-there'
    obj = DottedName(original_name)
    assert original_name.replace('-', '_') == obj.safe_name


def test_safe_name_all_at_once():
    original_name = 'Hi-There'
    obj = DottedName(original_name)
    assert obj.safe_name == 'hi_there'


def test_namespaces_none():
    obj = DottedName('plain_package_name')
    assert obj.namespaces == ['plain_package_name', ]


def test_namespaces_some():
    obj = DottedName('plone.app.dexterity')
    assert len(obj.namespaces) == 3


def test_namespaces_some_details():
    obj = DottedName('plone.app.dexterity')
    assert obj.namespaces == ['plone', 'app', 'dexterity', ]


def test_namespaces_really_long():
    obj = DottedName('one.two.three.four.five.six')
    assert len(obj.namespaces) == 6
    assert obj.namespaces == ['one', 'two', 'three', 'four', 'five', 'six', ]


def test_is_namespaced_not():
    obj = DottedName('nope')
    assert obj.is_namespaced is False


def test_is_namespaced():
    obj = DottedName('plone.app.dexterity')
    assert obj.is_namespaced


@pytest.mark.skipif(
    sys.version_info > (3, 0),
    reason='Fails on Python 3.x as it raises a TypeError',
)
def test_comparison_fallback():
    obj1 = DottedName('plone.app.dexterity')
    result = obj1 < object()
    assert result


def test_comparison_bigger_than():
    obj1 = DottedName('plone.app.ldap')
    obj2 = DottedName('plone.app.dexterity')
    assert obj1 > obj2


def test_comparison_bigger_or_equal_than():
    obj1 = DottedName('plone.app.ldap')
    obj2 = DottedName('plone.app.dexterity')
    assert obj1 >= obj2
    assert obj1 >= obj1


def test_comparison_smaller_than():
    obj1 = DottedName('plone.app.dexterity')
    obj2 = DottedName('plone.app.ldap')
    assert obj1 < obj2


def test_comparison_smaller_or_equal_than():
    obj1 = DottedName('plone.app.dexterity')
    obj2 = DottedName('plone.app.ldap')
    assert obj1 <= obj2
    assert obj1 <= obj1


def test_comparison_equality():
    obj1 = DottedName('plone.app.dexterity')
    assert obj1 == obj1


def test_comparison_equality_fallback():
    obj1 = DottedName('plone.app.dexterity')
    result = obj1 == object()
    assert not result


def test_repr():
    obj = DottedName('z3c.dependencychecker')
    assert repr(obj) == '<DottedName z3c.dependencychecker>'


def test_hash():
    obj1 = DottedName('bla')
    assert isinstance(hash(obj1), int)


def test_hash_in_use():
    obj1 = DottedName('one')
    obj2 = DottedName('two')
    obj3 = DottedName('three')
    obj4 = DottedName('one')
    uniques = {obj1, obj2, obj3, obj4, }
    assert len(uniques) == 3



# -*- coding: utf-8 -*-
from z3c.dependencychecker.package import Package
from z3c.dependencychecker.report import Report
from z3c.dependencychecker.tests.utils import write_source_file_at
import mock


def test_print_header_nothing(capsys):
    fake_package = mock.MagicMock()
    report = Report(fake_package)
    report._print_header(None)
    out, err = capsys.readouterr()
    assert out == ''


def test_print_header_text(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report._print_header('hi')
    out, err = capsys.readouterr()

    assert 'hi' in out
    assert '==' in out
    assert '===' not in out


def test_print_report(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_report()
    out, err = capsys.readouterr()
    assert '======' not in out  # there are no errors on the dependencies
    assert '' == err


def test_missing_requirements_nothing(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report.missing_requirements()
    out, err = capsys.readouterr()
    assert out == ''


def test_missing_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import this.package'
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n' \
           '====================' in out
    assert 'this.package' in out


def test_missing_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView'
    )
    write_source_file_at(
        (path, '{0}.egg-info'.format(package_name)),
        'requires.txt',
        'Zope2',
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'Zope2 = ["Four", "Products.Five", "Three" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n' \
           '====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView'
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'ignore-packages = ["Four", "Products.Five", "Three" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n' \
           '====================' not in out
    assert 'Products.Five' not in out
    assert 'Four' not in out
    assert 'Three' not in out


def test_missing_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import this.package',
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert 'Missing test requirements\n' \
           '=========================' in out
    assert 'this.package' in out


def test_missing_test_requirements_with_user_mapping(
    capsys,
    minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView'
    )
    write_source_file_at(
        (path, '{0}.egg-info'.format(package_name)),
        'requires.txt',
        'Zope2',
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'Zope2 = ["Products.Five" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n' \
           '====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_test_requirements_with_user_mapping_on_test_extra(
    capsys,
    minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView'
    )
    write_source_file_at(
        (path, '{0}.egg-info'.format(package_name)),
        'requires.txt',
        '\n'.join([
            '[test]',
            'Zope2',
        ]),
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'Zope2 = ["Products.Five" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n' \
           '====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_test_requirements_with_ignored_packages(
    capsys,
    minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView'
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'ignore-packages = ["Products.Five" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n' \
           '====================' not in out
    assert 'Products.Five' not in out


def test_unneeded_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import this.package',
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n' \
           '=====================' in out
    assert 'one' in out
    assert 'two' in out


def test_unneeded_requirements_with_ignored_packages(
        capsys,
        minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import this.package',
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'ignore-packages = ["one" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n' \
           '=====================' in out
    assert 'one' not in out
    assert 'two' in out


def test_unneeded_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join([
            '[test]',
            'pytest',
            'mock',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded test requirements\n' \
           '==========================' in out
    assert 'pytest' in out
    assert 'mock' in out


def test_unneeded_test_requirements_with_ignore_packages(
    capsys,
    minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join([
            '[test]',
            'pytest',
            'mock',
        ]),
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'ignore-packages = ["mock" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded test requirements\n' \
           '==========================' in out
    assert 'pytest' in out
    assert 'mock' not in out


def test_unneeded_test_requirements_no_tests_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out


def test_requirements_that_should_be_test_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import one',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import two',
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join([
            'one',
            'two',
            '[test]',
            'pytest',
            'mock',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert 'Requirements that should be test requirements\n' \
           '=============================================' in out
    assert 'two' in out


def test_requirements_that_should_be_test_requirements_with_ignored_packages(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import one',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        '\n'.join([
            'import two',
            'import three',
        ])
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join([
            'one',
            'two',
            'three',
            '[test]',
            'pytest',
            'mock',
        ]),
    )
    write_source_file_at(
        (path, ),
        'pyproject.toml',
        '\n'.join([
            '[tool.dependencychecker]',
            'ignore-packages = ["two" ]',
        ]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert 'Requirements that should be test requirements\n' \
           '=============================================' in out
    assert 'three' in out
    assert 'two' not in out


def test_print_notice(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_notice()
    out, err = capsys.readouterr()
    assert 'Note: ' in out
    assert '' == err


def test_exit_status_set(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.inspect()
    report = Report(package)
    assert report.exit_status == 0
    report.print_report()
    assert report.exit_status == 1



# -*- coding: utf-8 -*-
from z3c.dependencychecker.modules import PythonModule
from z3c.dependencychecker.tests.utils import write_source_file_at
import os
import pytest
import tempfile


IMPORT = 'import foo'
IMPORT_MULTIPLE = 'import foo, bar'
IMPORT_AS = 'import foo as bar'
IMPORT_AS_MULTIPLE = 'import foo as bar, boo as bla'

FROM_IMPORT = 'from foo import bar'
FROM_IMPORT_MULTIPLE = 'from foo import bar, ber'
FROM_IMPORT_AS = 'from foo import bar as ber'
FROM_IMPORT_AS_MULTIPLE = 'from foo import bar as ber, boo as bla'
FROM_IMPORT_ASTERISK = 'from os import *'
FROM_IMPORT_DOT = 'from . import something'
FROM_IMPORT_DOT_DOT = 'from .. import something'
FROM_IMPORT_DOT_RELATIVE = 'from .local import something'


def _get_imports_of_python_module(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath, ),
        source_code=source,
    )

    python_module = PythonModule(folder.strpath, temporal_file)
    dotted_names = [x.name for x in python_module.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = [x for x in PythonModule.create_from_files(path)]
    assert len(modules_found) == 0


def test_create_from_files_single_file():
    _, tmp_file = tempfile.mkstemp('.py')
    modules_found = [x for x in PythonModule.create_from_files(tmp_file)]
    assert len(modules_found) == 1
    assert modules_found[0].path == tmp_file


def test_create_from_files_no_init(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    assert len(os.listdir(src_path)) == 0

    write_source_file_at([src_path, ])
    assert len(os.listdir(src_path)) == 1

    modules_found = [x for x in PythonModule.create_from_files(src_path)]
    assert len(modules_found) == 0


def test_create_from_files_ignore_no_python(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at([src_path, ], filename='__init__.py')
    tempfile.mkstemp('.pt', 'bla', src_path)

    modules_found = [x for x in PythonModule.create_from_files(src_path)]
    assert len(modules_found) == 1


def test_create_from_files_found_python(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at([src_path, ], filename='__init__.py')
    write_source_file_at([src_path, ], filename='bla.py')

    modules_found = [x for x in PythonModule.create_from_files(src_path)]
    assert len(modules_found) == 2


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at([src_path, ], filename='__init__.py')
    write_source_file_at([src_path, ], filename='bla.py')
    write_source_file_at([src_path, 'a', ], filename='__init__.py')
    write_source_file_at([src_path, 'a', ], filename='bla.py')
    write_source_file_at([src_path, 'a', 'b', ], filename='__init__.py')
    write_source_file_at([src_path, 'a', 'b', ], filename='bla.py')

    modules_found = [x for x in PythonModule.create_from_files(src_path)]
    assert len(modules_found) == 6


def test_import(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT)
    assert len(dotted_names) == 1


def test_import_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT)
    assert dotted_names == ['foo', ]


def test_import_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_MULTIPLE)
    assert len(dotted_names) == 2


def test_import_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_MULTIPLE)
    assert sorted(dotted_names) == ['bar', 'foo', ]


def test_import_as(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS)
    assert len(dotted_names) == 1


def test_import_as_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS)
    assert dotted_names == ['foo', ]


def test_import_as_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS_MULTIPLE)
    assert len(dotted_names) == 2


def test_import_as_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS_MULTIPLE)
    assert sorted(dotted_names) == ['boo', 'foo', ]


def test_from_import(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT)
    assert len(dotted_names) == 1


def test_from_import_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT)
    assert dotted_names == ['foo.bar', ]


def test_from_import_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_MULTIPLE)
    assert len(dotted_names) == 2


def test_from_import_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_MULTIPLE)
    assert sorted(dotted_names) == ['foo.bar', 'foo.ber', ]


def test_from_import_as(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_AS)
    assert len(dotted_names) == 1


def test_from_import_as_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_AS)
    assert dotted_names == ['foo.bar', ]


def test_import_asterisk(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_ASTERISK)
    assert len(dotted_names) == 1


def test_import_asterisk_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_ASTERISK)
    assert dotted_names == ['os', ]


@pytest.mark.parametrize(
    'statement',
    [
        FROM_IMPORT_DOT,
        FROM_IMPORT_DOT_DOT,
        FROM_IMPORT_DOT_RELATIVE,
    ]
)
def test_ignore_local_imports(statement, tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, statement)
    assert len(dotted_names) == 0


def test_from_import_as_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        FROM_IMPORT_AS_MULTIPLE,
    )
    assert len(dotted_names) == 2


def test_from_import_as_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        FROM_IMPORT_AS_MULTIPLE,
    )
    assert sorted(dotted_names) == ['foo.bar', 'foo.boo', ]


def test_imports_multiple_lines(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        'import foo\nimport bar',
    )
    assert sorted(dotted_names) == ['bar', 'foo', ]



from setuptools import find_packages
from setuptools import setup
import codecs
import os.path


version = '2.5.1'


def read(filename):
    try:
        with codecs.open(filename, encoding='utf-8') as f:
            return unicode(f.read())
    except NameError:
        with open(filename, 'r', encoding='utf-8') as f:
            return f.read()



long_description = u'\n\n'.join([
    read('README.rst'),
    read(os.path.join('z3c', 'dependencychecker', 'USAGE.rst')),
    read('CHANGES.rst'),
    ])


setup(
    name='z3c.dependencychecker',
    version=version,
    description="Reports on missing or unneeded setup.py dependencies",
    long_description=long_description,
    long_description_content_type='text/x-rst; charset=UTF-8',
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Quality Assurance',
    ],
    keywords='dependencies requirements missing imports',
    author='Reinout van Rees',
    author_email='reinout@vanrees.org',
    url='https://github.com/reinout/z3c.dependencychecker',
    license='BSD',
    packages=find_packages(exclude=['ez_setup', ]),
    namespace_packages=['z3c', ],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'stdlib-list',
        'cached-property',
        'toml',
    ],
    extras_require={
        'test': [
            'pytest',
            'mock',
        ],
    },
    entry_points={
        'console_scripts': [
            'dependencychecker = z3c.dependencychecker.main:main',
        ]
    },
)



# -*- coding: utf-8 -*-
from z3c.dependencychecker.modules import DocFiles
from z3c.dependencychecker.tests.utils import write_source_file_at

import os


NO_DOC = '\n'.join([
    'Random title',
    '============',
    '',
    'Random paragraph',
])
INVALID_PYTHON = '\n'.join([
    'Random title',
    '============',
    '',
    'Random paragraph',
    '>>> and some text here that breaks the parser'
])
SINGLE_IMPORT = '\n'.join([
    'Random title',
    '============',
    '',
    'Random paragraph',
    '>>> import zope.annotation'
])
MULTIPLE_IMPORTS_SAME_LINE = '\n'.join([
    'Random title',
    '============',
    '',
    'Random paragraph',
    '>>> from zope.interface import Interface, Attribute',
])
MULTIPLE_IMPORTS_DIFFERENT_LINES = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> from zope.component import adapter',
    '        >>> from zope.component import utility',
    '        """',
])
MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> from zope.component import adapter',
    '        >>> this should  be skipped and next line processed happily',
    '        >>> from zope.component import utility',
    '        """',
])


def _get_dependencies_on_file(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath, ),
        source_code=source,
    )

    doc_file = DocFiles(folder.strpath, temporal_file)
    dotted_names = [x.name for x in doc_file.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    modules_found = [x for x in DocFiles.create_from_files(src_path)]
    assert len(modules_found) == 0


def test_create_from_files_deep_nested_txt(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at(
        [src_path, 'a', 'b', 'c', ],
        filename='bla.txt',
    )

    modules_found = [x for x in DocFiles.create_from_files(src_path)]
    assert len(modules_found) == 1


def test_create_from_files_deep_nested_rst(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at(
        [src_path, 'a', 'b', 'c', ],
        filename='bla.rst',
    )

    modules_found = [x for x in DocFiles.create_from_files(src_path)]
    assert len(modules_found) == 1


def test_no_code(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, NO_DOC)
    assert len(dotted_names) == 0


def test_invalid_code_ignored(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, INVALID_PYTHON)
    assert len(dotted_names) == 0


def test_code_found(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, SINGLE_IMPORT)
    assert len(dotted_names) == 1


def test_code_found_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, SINGLE_IMPORT)
    assert dotted_names == ['zope.annotation', ]


def test_always_testing(tmpdir):
    temporal_file = write_source_file_at(
        (tmpdir.strpath, ),
        source_code='>>> import foo',
        filename='file.rst'
    )

    doc_file = DocFiles(tmpdir.strpath, temporal_file)
    dotted_names = [x for x in doc_file.scan()]
    assert len(dotted_names) == 1
    assert dotted_names[0].is_test


def test_multiple_imports_same_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_same_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert 'zope.interface.Interface' in dotted_names
    assert 'zope.interface.Attribute' in dotted_names


def test_multiple_imports_different_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_different_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )

    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names


def test_multiple_imports_with_invalid_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_with_invalid_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )

    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names



# -*- coding: utf-8 -*-
from z3c.dependencychecker.modules import PythonDocstrings
from z3c.dependencychecker.tests.utils import write_source_file_at


NO_DOC = '\n'.join([
    'class MyClass(object):',
    '    pass',
])
INVALID_PYTHON = '\n'.join([
    '"""',
    '>>> and some text here that breaks the parser'
    '"""',
])
DOC_IN_MODULE = '\n'.join([
    '"""',
    'Random docstring with code to be evaluated.',
    '',
    '>>> import zope.component',
    '"""',
])
DOC_IN_FUNCTION = '\n'.join([
    'def test():'
    '    """Random docstring with code to be evaluated.',
    '',
    '    >>> import zope.component.adapter',
    '    """',
])
DOC_IN_CLASS = '\n'.join([
    'class MyClass(object):'
    '    """Random docstring with code to be evaluated.'
    '',
    '    >>> from zope.component import utility',
    '    """',
])
DOC_IN_METHOD = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> import zope.interface',
    '        """',
])
MULTIPLE_IMPORTS_SAME_LINE = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> from zope.interface import Interface, Attribute',
    '        """',
])
MULTIPLE_IMPORTS_DIFFERENT_LINES = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> from zope.component import adapter',
    '        >>> from zope.component import utility',
    '        """',
])
MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN = '\n'.join([
    'class MyClass(object):',
    '    def test(self):',
    '        """Docstring with code to be evaluated.',
    '',
    '        >>> from zope.component import adapter',
    '        >>> this should  be skipped and next line processed happily',
    '        >>> from zope.component import utility',
    '        """',
])


def _get_dependencies_on_file(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath, ),
        source_code=source,
    )

    docstring = PythonDocstrings(folder.strpath, temporal_file)
    dotted_names = [x.name for x in docstring.scan()]
    return dotted_names


def test_no_docstring(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, NO_DOC)
    assert len(dotted_names) == 0


def test_invalid_code_ignored(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, INVALID_PYTHON)
    assert len(dotted_names) == 0


def test_docstring_in_module(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_MODULE)
    assert len(dotted_names) == 1


def test_docstring_in_module_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_MODULE)
    assert dotted_names == ['zope.component', ]


def test_docstring_in_function(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_FUNCTION)
    assert len(dotted_names) == 1


def test_docstring_in_function_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_FUNCTION)
    assert dotted_names == ['zope.component.adapter', ]


def test_docstring_in_class(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_CLASS)
    assert len(dotted_names) == 1


def test_docstring_in_class_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_CLASS)
    assert dotted_names == ['zope.component.utility', ]


def test_docstring_in_method(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_METHOD)
    assert len(dotted_names) == 1


def test_docstring_in_method_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_METHOD)
    assert dotted_names == ['zope.interface', ]


def test_docstring_multiple_imports_same_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_same_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert 'zope.interface.Interface' in dotted_names
    assert 'zope.interface.Attribute' in dotted_names


def test_docstring_multiple_imports_different_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_same_different_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names


def test_docstring_multiple_imports_with_invalid_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_with_invalid_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names
